/******************************************************************************
*
* project name:  TICT-Explorer
* file name:     common.h
* initial date:  20/01/2001
* author:        thomas.nussbaumer@gmx.net
*
* description:   common definitions used in more than one file
*
******************************************************************************/

//#define MSGS_EXPLORER
//#include "messages.h"

#if !defined(TICTEX_COMMON_H)
#define TICTEX_COMMON_H

//-----------------------------------------------------------------------------
// check for -DUSE_TI89 commandline flag (for TI-89 version)
// *** or ***
// check for -DUSE_TI92PLUS commandline flag (for TI-92p/V200 version)
//
// ONLY ONE IS ALLOWED PER COMPILE RUN
//-----------------------------------------------------------------------------
#if defined(USE_TI89) && defined(USE_TI92PLUS)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92PLUS)
#undef USE_V200
#define USE_V200           // treat V200 as TI92p
#define C89_92(x,y) (y)
#else
#error Please define either USE_TI89 or USE_TI92PLUS (V200 will be treated as TI92p)
#endif

//-----------------------------------------------------------------------------
// number of maximum characters used for a file comment
//-----------------------------------------------------------------------------
#define MAX_COMMENT 31

//-----------------------------------------------------------------------------
// special structure similar to SYM_ENTRY, but with some nice extensions like
// 0-ending strings (easier to print ;-)
//-----------------------------------------------------------------------------
typedef struct {
    unsigned char  name[9];
    unsigned char  type;
    unsigned char  ext[5];
    unsigned char  marked;
    unsigned char  size[6];
    unsigned char  comment[MAX_COMMENT+1];

    union
      {
        unsigned short flags_n;
        struct
          {
            unsigned short busy      : 1, local       : 1, flag1_5     : 1, flag1_4     : 1,
                           collapsed : 1, twin        : 1, archived    : 1, in_view     : 1,
                           folder    : 1, overwritten : 1, checked     : 1, hidden      : 1,
                           locked    : 1, statvar     : 1, graph_ref_1 : 1, graph_ref_0 : 1;
          } bits;
      } flags;
} file_t;

//-----------------------------------------------------------------------------
// special structure used for folder entries (just the name ...)
//-----------------------------------------------------------------------------
typedef struct {
    unsigned char name[9];
} folder_t;

//-----------------------------------------------------------------------------
// definitions of mode (where we are: in the folder list or in the file list)
// and the global which is used to store the active mode
//-----------------------------------------------------------------------------
#define SCROLL_FOLDER 0
#define SCROLL_FILE   1

//-----------------------------------------------------------------------------
// calculator depended definitions and globals
//-----------------------------------------------------------------------------
#define PIX_WIDTH   C89_92(160,240)
#define PIX_HEIGHT  C89_92(100,128)
#define OFFSET_X    C89_92(0,40)
#define OFFSET_Y    C89_92(0,14)
#define KEY_UP2ND   C89_92(4433,4434)
#define KEY_DOWN2ND C89_92(4436,4440)

#define VAL_DIAMOND C89_92(16384,8192)
#define VAL_SHIFT   C89_92(8192,16384)

#define NR1_2ND   C89_92(34,  149)
#define NR2_2ND   C89_92(92,  4146)
#define NR3_2ND   C89_92(4147,4147)
#define NR4_2ND   C89_92(58,  4148)
#define NR5_2ND   C89_92(4149,4149)
#define NR6_2ND   C89_92(4150,4150)
#define NR7_2ND   C89_92(4151,4151)
#define NR8_2ND   C89_92(4152,4152)
#define NR9_2ND   C89_92(59,  4153)

#include "messages.h"

#define MSG_OKAY                   0
#define MSG_PRGTITLE               1
#define MSG_ERR_CRASHOFFSET        2

#define MSG_ERR_WRONGVERSION      18
#define MSG_ERR_OUTOFMEM          19
#define MSG_ERR_LOCKINGFAILED     20
#define MSG_ERR_PPGINVALID        21
#define MSG_ERR_EXECFAILED        22
#define MSG_ERR_NOFOLDERS         23
#define MSG_ERR_FILEINUSE         24
#define MSG_ERR_FILENOTFOUND      25
#define MSG_ERR_INVALIDEXPLLIB    26
#define MSG_ERR_EXPLLIBNOTFOUND   27
#define MSG_ERR_TWINFAILED        28
#define MSG_ERR_TWINDEL           29

#if defined(MSGS_EXPLORER)
#define MSG_ERR_EBOOKNOTFOUND      30
#define MSG_ERR_QUILLNOTFOUND      31
#define MSG_ERR_SCOTTNOTFOUND      32
#define MSG_ERR_PVVIEWNOTFOUND     33
#define MSG_ERR_PRGNOTALLOWED      34
#else
#define MSG_ERR_FINDMYSELF         30
#endif

#endif

//=============================================================================
// Revision History
//=============================================================================
//
