/******************************************************************************
*
* project name:  TICT-Explorer
* file name:     tictex.c
* initial date:  19/12/2000
* author:        thomas.nussbaumer@gmx.net
*
* description:   special PPG launcher for TICT-Explorer
*
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*
* !!! IMPORTANT WARNING !!!
*
*
* Don't modify anything nor recompile this program with any other TIGCC Library
* than 2.41 and TIGCC 0.93, BEFORE you have completely understood the
* mechanism of the stack backup. Otherwise the TICT-Explorer won't run stable.
*
* Please use batchfile build.bat to rebuild !!!
*
* The STACK_CORRECT_VALUE is evaluated by running tool tststack() twice:
* (1) one time from the commandline
* (2) one time from the explorer
*
* The difference of the two values is used as STACK_CORRECT_VALUE
*
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*
* $Id: tictex.c,v 1.17 2002/10/29 09:07:25 tnussb Exp $
*
******************************************************************************/

//-----------------------------------------------------------------------------
// check for -DUSE_TI89 commandline flag (for TI-89 version)
// *** or ***
// check for -DUSE_TI92PLUS commandline flag (for TI-92p version)
//
// ONLY ONE IS ALLOWED PER COMPILE RUN
//-----------------------------------------------------------------------------
#if defined(USE_TI89) && defined(USE_TI92PLUS)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92PLUS)
#undef USE_V200            // No warnings please.
#define USE_V200           // V200 support !!!
#define C89_92(x,y) (y)
#else
#error Please define either USE_TI89 or USE_TI92PLUS (V200 will be treated as TI92+)
#endif

#define SET_FILE_IN_USE_BIT
#include <tigcclib.h>

#include "common.h"
#include "interface.h"
#define MINIMIZE_VATUTILS
#include "vatutils.h"
#include "protect.h"
#include "tts/ttunpack.h"

extern void HomeScreenRestore(void);


//#define CAUSE_CRASH          // when you uncomment this a crash will be produced
// Uncomment to use the launch method that creates twin symbols. Warning, this does
// not suit to kernel-based programs under PreOS 0.7+...
//#define OLD_FASHIONED_LAUNCH

//-----------------------------------------------------------------------------
// to evaluate the correct STACK_CORRECT_VALUE set it to 0 and run tststack()
// twice: one time from the commandline and one time from the explorer.
// Then set CORRECT_VALUE to the difference of both runs.
// TICT-Explorer 1.40: I found 84 - exactly half - and I wonder why, so I'm
// leaving 168.
//-----------------------------------------------------------------------------
#define STACK_CORRECT_VALUE (168)
#define STACK_MAX_SIZE (220+STACK_CORRECT_VALUE)
unsigned char* stack_buffer = NULL;
unsigned char* stack_orig   = NULL;
unsigned char* stack_tmp    = NULL;

#if 0
char st_show_buf[30] = {0};
unsigned long st_show_val = 0;
#define SHOW_STACK(a) ({\
   asm volatile ("move.l %sp,st_show_val");\
   sprintf(st_show_buf,"ST%d=0x%08lX (press a key)",a,st_show_val);\
   ST_helpMsg(st_show_buf);ngetchx();})

#else
#define SHOW_STACK(a)
#endif

//----------------------------------------------------------------------
// sophisticated macro to give started program as much stack as possible
//
// Before TICT-Explorer 1.40, there used to be no indication on
// clobbered registers.
//----------------------------------------------------------------------
#define SAVE_STACK() ({ \
    /*__HALT;*/\
    asm volatile("movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)\n"\
                 "move.l %%sp,stack_tmp\n"\
                 "move.l stack_tmp,%%d0\n"\
                 "move.l stack_orig,%%d1\n"\
                 "move.l %%d0,%%a0\n"\
                 "sub.l %%d0,%%d1\n"\
                 "move.l stack_buffer,%%a1\n"\
                 "0: move.b  (%%a0),(%%a1)+\n"\
                 "move.b  0,(%%a0)+\n"\
                 "dbra %%d1,0b\n"\
                 "move.l stack_orig,%%sp" : : : "d0","d1","d2","a0","a1","cc","memory");})

//----------------------------------------------------------------------
// restores stack state saved with SAVE_STACK
//
// Before TICT-Explorer 1.40, there used to be no indication on
// clobbered registers. I know this one is wrong
// (d0-d7/a0-a6 are destroyed), but it should be better than nothing.
//----------------------------------------------------------------------
#define RESTORE_STACK() ({ \
    /*__HALT;*/\
    asm volatile("move.l stack_tmp,%%d0\n"\
                 "move.l stack_orig,%%d1\n"\
                 "move.l %%d0,%%a0\n"\
                 "sub.l %%d0,%%d1\n"\
                 "move.l stack_buffer,%%a1\n"\
                 "0:move.b  (%%a1)+,(%%a0)+\n"\
                 "dbra %%d1,0b\n"\
                 "move.l stack_tmp,%%sp\n"\
                 "movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6" : : : "d0","d1","d2","a0","a1","cc","memory");})
//                 "movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6" : : : "cc","memory");})



volatile TICTEX_INTERFACE interface = EMPTY_INTERFACE_INIT;
unsigned char* patch_addr           = NULL;
ESI            top_backup           = 0;
char           sbuffer[40]          = {};
char           tmpname[30]          = {};
char           initial_folder[30]   = {};
//char*          lcdbuffer            = NULL;
char           explorer_folder[9]   = {};

//---------------------------------------------------------------------
// NOTE: the NOPs are very important due to a hack with RETURN_VALUE
//       (they must be exactly after the call!!)
//---------------------------------------------------------------------
#define ASM_NOP_CALL(x) ({\
asm volatile("movem.l %d0-%d7/%a0-%a6,-(%sp)"); \
ASM_fastcall(x);\
asm volatile("nop;nop;nop;nop;");\
asm volatile("movem.l (%sp)+,%d0-%d7/%a0-%a6");})


PROTECTION_T   temporary_protection;
ERROR_FRAME    temporary_errorframe;

PROTECTION_T   main_protection;
ERROR_FRAME    main_errorframe;


// Launching code taken from the AMS native fork of SIDE.
// Used with Kevin's permission, original comments left.
// Modified by Lionel to make an optimization and add crash protection.
void launch(void *cptroff asm("a3"), void **percatcherpc asm("a2"));

asm("
| LAUNCH DECOMPRESSED PROGRAM
| NOTE: Some of this code is the same as in ttstart. This is
|       because this is code _I_ contributed to ttstart. So
|       it is being relicensed by the original author.

launch:
movem.l %d1-%d7/%a0-%a6,-(%sp)
moveq.l #0,%d6
move.w (%a3)+,%d6
| Round the length up to an even size
| NOTE: The handle size is rounded up automatically.
addq.l #1,%d6
moveq.l #0xfffffffe,%d0
and.l %d0,%d6
| Save error catcher PC
lea.l ercatcherpc(%pc),%a0
move.l (%a2),(%a0)
| Set new error catcher PC
lea.l 0(%a3,%d6.l),%a0
lea.l (ercatcher-middle)(%a0),%a1
move.l %a1,(%a2)
| Append the stub to the program
lea.l middle(%pc),%a1
moveq.l #50,%d0
stub_copy_loop:
move.w (%a1)+,(%a0)+
dbra.w %d0,stub_copy_loop
| Unprotect and run the program
pea target(%PC)
movea.l %a3,%a4
pea (%a3,%d6.l)
jbra unprotect_a3_jumpto_a4
target:
| Clean up
lea (%sp,36),%sp
move.l (%sp)+,%d0
movem.l (%sp)+,%d1-%d7/%a0-%a6
rts


| CODE EXECUTED AT END OF PROGRAM
| NOTE: This is the same code as in ttstart, because the code in ttstart
|       is actually MY code.

middle:
nop
nop
nop
| Unprotect and jump to the launcher
movea.l (%sp)+,%a4
lea __ld_entry_point-target(%a4),%a3
move.l %d0,16(%sp)
| Unprotection routine
unprotect_a3_jumpto_a4:
cmp.l #0x40000,%a3
bcs 0f
lea.l 0x3f000,%a3
0:
lea (%sp,-20),%sp
move.l 0xc8.w,%d0
andi.l #0xE00000,%d0
addi.l #0x20000,%d0
move.l %d0,(%sp,12)
move.l %d0,(%sp,16)
move.l 20(%sp),(%sp)
move.l 24(%sp),4(%sp)
trap #0xC
move.w #0x2700,%sr
moveq.l #0xF,%d3
pea (%a4)
clr.w  -(%sp)
move.l 0xAC,%a0
jmp (%a0)
ercatcherentrypoint: .long __ld_entry_point
ercatcherpc: .long 0
.even
ercatcher:
movem.l %d0-%d7/%a0-%a6,-(%sp)
move.l ercatcherentrypoint(%PC),%a3
lea.l ercatchercleanup-__ld_entry_point(%a3),%a4
jbra unprotect_a3_jumpto_a4


| LAUNCHING CLEANUP CODE EXECUTED IN THE LAUNCHER'S EXECUTION SPACE

ercatchercleanup:
lea (%sp,20),%sp
movem.l (%sp)+,%d0-%d7/%a0-%a6
move.l ercatcherpc(%PC),-(%a7)
rts
");


/*===========================================================================*/
/* executes a block of memory in a TRY/CATCH block (secure)                  */
/* if execution failed exe_okay will be set to 0 otherwise 1                 */
/*===========================================================================*/
void ExecuteSecure(unsigned char* addr,unsigned char* reloc_end) {
    unsigned long offset = (~*(long*)0xc8 & 0x800000) >> 5;
    void *oldlastexec;

    FontSetSys(F_6x8);
    EX_patch(addr + offset,reloc_end + offset);

    //asm("0: bra.s 0b");
    // Code taken from ttstart.
    InstallCrashProtection(&temporary_protection);
    if (!ER_catch(temporary_errorframe)) {
        #if defined(DEBUG_HALT_BEFORE)
        __HALT;
        #endif
        SHOW_STACK(1);

        // Fix "APD crash" if necessary and run program.
        if (TIOS_entries < 1000) {
//        if (AMS_1xx) {
            TRY
                asm("
                movem.l %%d0-%%d7/%%a0-%%a6,-(%%sp)
                jsr (%0)
                nop
                nop
                nop
                movem.l (%%sp)+,%%d0-%%d7/%%a0-%%a6
                "::"a"(addr):"cc");
          FINALLY
//           error(ERR_EXEC);
          ENDFINAL
        }
        else {
            char *p;// = (char *)ROM_base;// + 0x12000;
            long i;
            asm volatile("
            move.l 0xC8.w,%0
            andi.l #0xE00000,%0
            addi.l #0x12000,%0
            move.l %0,%1
            move.w #0x6000/2-1,%0 | Set up iteration counter
            0:
            cmpi.l #0xDEADDEAD,(%1) | Look for the stack fence
            addq.l #2,%1 | Increment pointer, condition codes unchanged
            dbeq %0,0b
            " : "=d"(i), "=a"(p) : : "d0","a0","cc");
            p+=1[(short *)p]?6:10;
            oldlastexec=*(void **)(long)*(short *)p;
            *(void **)(long)*(short *)p=offset?(void*)0x3f000:(addr+2);
            TRY
//                asm("0: bra.s 0b");
                launch(addr + offset - 2, (void**)&__errFrame->PC);
            FINALLY
                //asm("0: bra.s 0b");
            ENDFINAL
            *(void **)(long)*(short *)p = oldlastexec;
        }

        ER_success();
    }

    #if defined(DEBUG_HALT_AFTER)
    __HALT;
    #endif
    RemoveCrashProtection(&temporary_protection);
}


/*===========================================================================*/
/* try to execute a PPG file                                                 */
/*                                                                           */
/* returns MSG_OKAY if everything is okay                                    */
/*===========================================================================*/
static inline short StartPPG(HANDLE handle) {
    unsigned char*   src;
    unsigned char*   dest;
    unsigned short   length;
    unsigned short   retval;
    HANDLE           uncomp_h;

    if (!(src=SaveHeapLock(handle))) return MSG_ERR_LOCKINGFAILED;

    length = *(unsigned short*)(src)+2;

    src+=2;

    if (memcmp(src+length-7,"ppg\x00\xF8",5) && !ttunpack_valid(src)) {
        SaveHeapUnLock();
        return MSG_ERR_PPGINVALID;
    }

    length = ttunpack_size(src);

    // NOTE: we don't use malloc, but the low level routines
    //       otherwise compressed kernel-dependent progs will crash
    if (!(uncomp_h=HeapAllocHigh(length+120))) { // +120 bytes for Titanium support.
        SaveHeapUnLock();
        return MSG_ERR_OUTOFMEM;
    }
    dest = HeapDeref(uncomp_h);

    //-------- OLD BLOCK - replaced by above one!
    //if (!(dest=malloc(length))) {     // allocate memory for original
    //    SaveHeapUnLock();
    //    return MSG_ERR_OUTOFMEM;
    //}

    ST_helpMsg("decompressing ...");
    retval = ttunpack_decompress(src,dest);
    ST_eraseHelp();

    //------------------------------
    // "houston we got problems ..."
    //------------------------------
    if (retval) {
        HeapFree(uncomp_h); // free(dest); -- using now HeapFree() for kernel-ppg support
        SaveHeapUnLock();
        return MSG_ERR_PPGINVALID;
    }

    //-------------------------------------------------------------------------
    // relocate program and start it
    //-------------------------------------------------------------------------
    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !
    SAVE_STACK();  // MUST BE DONE OUTSIDE DUE TO ER_CATCH() !!!!
    ExecuteSecure(dest+2,dest+length-1);
    RESTORE_STACK();
    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !

    HeapFree(uncomp_h); //free(dest); -- using now HeapFree() for kernel-ppg support
    SaveHeapUnLock();

    if (EXCEPTION_TRIGGERED()) {
        retval = GET_EXCEPTION_CODE()+MSG_ERR_CRASHOFFSET;
        RESET_EXCEPTION();
        return retval;
    }
    else {
        return MSG_OKAY;
    }
}


/*===========================================================================*/
/* try to patch the interface address into the explorer                      */
/* returns 1 if okay                                                         */
/*===========================================================================*/
static inline short PatchInterfaceAddress(unsigned char* cptr,unsigned short len) {
    unsigned char* endptr;

    //cptr  += 0x40000; // necessary? I don't think so ...
    endptr = cptr + ((len+1)&0xFFFE);
//    endptr = cptr + len - len % 2;

    patch_addr = NULL;

    do {
        if (*(unsigned long*)cptr      == MAGIC_EXPLORER1 &&
            *(unsigned long*)(cptr+4)  == MAGIC_EXPLORER2)
/*
        if (*(unsigned short*)cptr     == MAGIC_EXPLORER14 &&
            *(unsigned short*)(cptr+2) == MAGIC_EXPLORER24 &&
            *(unsigned short*)(cptr+4) == MAGIC_EXPLORER34 &&
            *(unsigned short*)(cptr+6) == MAGIC_EXPLORER44)
*/
        {
#if defined(CAUSE_CRASH)
            *endptr=*cptr/0;
#endif

            (*(unsigned long*)cptr)     = (unsigned long)&interface;
            (*(unsigned long*)(cptr+4)) = (unsigned long)TICTEX_VERSION;
            patch_addr = cptr;

            return 1;
        }
        cptr+=2;
    }
    while (cptr < endptr);

    return 0;
}


/*===========================================================================*/
/* try to start an executable                                                */
/* returns MSG_OKAY if everything is okay                                    */
/*===========================================================================*/
#ifndef OLD_FASHIONED_LAUNCH
// macro which exchanges the content of two variables using the ASM exg instruction
#define ASM_XCHG(a,b)    asm volatile ("exg %0,%1" : "=da" (a), "=da" (b) : "0" (a), "1" (b) : "cc")

short StartExecutable(SYM_ENTRY* symptr, short is_explorer) {
    char*           cptr;
    unsigned short  plen;
    HANDLE          h; // Handle of executable.
    HANDLE          h2; // Handle of copy in case it is necessary.
    unsigned char   twin; // Boolean indicating whether a (false) twin exists.
    unsigned char  *ptr;

    //asm("0: bra.s 0b");
    h = symptr->handle;
    h2 = h;
    twin = 0;
    plen = HeapSize(h);

    if (symptr->flags.bits.archived) {
        h2 = HeapAlloc(plen);
        if (!h2) return MSG_ERR_TWINFAILED;
        ASM_XCHG(h,h2);
        //h ^= h2 ^= h ^= h2; // Exchange h, h2.
        twin = 0xFF;
    }

    // We must grow the handle due to Titanium support, which adds
    // code at the end of the program... As you can guess, not
    // doing so results in a crash (infinite loop in HeapAlloc <=>
    // memory corruption, obviously).
    // This might break programs which are very close to the
    // size limit (65518 bytes). A too large size should in practice
    // be the only reason why the HeapRealloc would fail, as
    // EM_twinSymFromExtMem does not seem to lock the handle it
    // creates, although it may also be due to a heavily fragmented
    // RAM with too many locked blocks. HeapCompress() is called
    // before attempting to enlarge the handle, in order to maximize
    // the probability of success.
    // I had better making an new message to describe the error
    // accurately, but returning MSG_ERR_TWINFAILED makes some
    // sense.
    HeapCompress();
    if (!(HeapRealloc(h,plen+120))) return MSG_ERR_TWINFAILED;

    if (!(cptr = SaveHeapLock(h))) return MSG_ERR_LOCKINGFAILED;

    if (twin) {
        // This is a false twin.
        // Need a memcpy !!
        memcpy(cptr,HeapDeref(h2),plen);
    }

    // The explorer is embedded into another file.
    if (is_explorer) cptr+=2;

    plen = *(unsigned short*)(cptr) + 2;

    if (is_explorer) {
        if (!PatchInterfaceAddress(cptr,plen)) {
            SaveHeapUnLock();
            if (twin) {
                HeapFree(h);
            }
            return MSG_ERR_INVALIDEXPLLIB;
        }
//        if (twin) {
//            plen -= 5; // Remove size of 0x00 L I B 00 0xF8 to point to 0xF3.
//        }
    }

    ptr = cptr+plen;

    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !
    SAVE_STACK();  // MUST BE DONE OUTSIDE DUE TO ER_CATCH() !!!!
    ExecuteSecure(cptr+2,ptr-1);
    RESTORE_STACK();
    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !

    if (is_explorer) {
        (*(unsigned long*)patch_addr)     = (unsigned long)MAGIC_EXPLORER1;
        (*(unsigned long*)(patch_addr+4)) = (unsigned long)MAGIC_EXPLORER2;

        // !! DIRTY-DIRTY-DIRTY !!
        // Restore 00 L I B 00 OTH_TAG, which was destroyed by ExecuteSecure.
        // This is simpler and less error-prone than modifying the launching code !
//asm("0: bra.s 0b");
        //        asm volatile("move.l %0,%%d0;add.l %1,%%d0;0: bra.s 0b" : : "g"(cptr), "g"(plen) : "d0","cc");

        *ptr++ = 0x00;
        *ptr++ = 'L';
        *ptr++ = 'I';
        *ptr++ = 'B';
        *ptr++ = 0x00;
        *ptr++ = OTH_TAG;
    }

    SaveHeapUnLock();

    if (twin) {
        HeapFree(h);
    }
    else {
        if (!(HeapRealloc(h,HeapSize(h)-120))) return MSG_ERR_TWINDEL;
    }

    if (EXCEPTION_TRIGGERED()) {
        plen = GET_EXCEPTION_CODE()+MSG_ERR_CRASHOFFSET;
        RESET_EXCEPTION();
        return plen;
    }
    else {
        return MSG_OKAY;
    }
}

#else
#warning Using old-fashioned launch !
short StartExecutable(SYM_ENTRY* symptr, short is_explorer) {
    char*           cptr;
    unsigned short  plen;
    HANDLE          h;
    HSym            twin = HS_NULL;
    char*           tmpptr = (char*)interface.foldername;
    SYM_ENTRY*      se;
    unsigned char *ptr;

    h = symptr->handle;

    if (symptr->flags.bits.archived) {
        strncpy(tmpname,symptr->name,8);
        tmpname[8]=0;
        if (is_explorer) tmpptr = explorer_folder;
        sprintf(sbuffer,"%s\\%s",tmpptr,tmpname);
        twin = EM_twinSymFromExtMem(GenerateTIOSName(sbuffer,tmpname),HS_NULL);
        if (!twin.folder) return MSG_ERR_TWINFAILED;
        se = DerefSym(twin);

        // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        // we set the hidden flag on TWIN symbols, otherwise
        // it may get deleted too early by a well-known AMS bug!
        // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        se->flags.bits.hidden = 1;
        h = se->handle;
    }

    // We must grow the handle due to Titanium support, which adds
    // code at the end of the program... As you can guess, not
    // doing so results in a crash (infinite loop in HeapAlloc <=>
    // memory corruption, obviously).
    // This might break programs which are very close to the
    // size limit (65518 bytes). A too large size should in practice
    // be the only reason why the HeapRealloc would fail, as
    // EM_twinSymFromExtMem does not seem to lock the handle it
    // creates, although it may also be due to a heavily fragmented
    // RAM with too many locked blocks.
    // I had better making an new message to describe the error
    // accurately, but returning MSG_ERR_TWINFAILED makes some
    // sense.
    HeapCompress();
    if (!(HeapRealloc(h,HeapSize(h)+120))) return MSG_ERR_TWINFAILED;

    if (!(cptr = SaveHeapLock(h))) return MSG_ERR_LOCKINGFAILED;

    // The explorer is embedded into another file.
    if (is_explorer) cptr+=2;

    plen = *(unsigned short*)(cptr) + 2;

    if (is_explorer) {
        if (!PatchInterfaceAddress(cptr,plen)) {
            SaveHeapUnLock();
            if (twin.folder) {
                twin = SymFind(GenerateTIOSName(sbuffer,tmpname));
                if (!twin.folder) return MSG_ERR_TWINDEL;
                se = DerefSym(twin);
                if (!(se->flags.bits.twin)) return MSG_ERR_TWINDEL;
                se->flags.bits.hidden = 0;
                if (!SymDelTwin(se)) return MSG_ERR_TWINDEL;
            }
            return MSG_ERR_INVALIDEXPLLIB;
        }
    }

    ptr = cptr+plen;

    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !
    SAVE_STACK();  // MUST BE DONE OUTSIDE DUE TO ER_CATCH() !!!!
    ExecuteSecure(cptr+2,ptr-1);
    RESTORE_STACK();
    // WARNING: *NO* ASSIGNMENTS TO d3-d7/a2-a6 BETWEEN SAVE_STACK and RESTORE_STACK !

    if (is_explorer) {
        (*(unsigned long*)patch_addr)     = (unsigned long)MAGIC_EXPLORER1;
        (*(unsigned long*)(patch_addr+4)) = (unsigned long)MAGIC_EXPLORER2;

        // !! DIRTY-DIRTY-DIRTY !!
        // Restore 00 L I B 00 OTH_TAG, which was destroyed by ExecuteSecure.
        // This is simpler and less error-prone than modifying the launching code !
//asm("0: bra.s 0b");
        //        asm volatile("move.l %0,%%d0;add.l %1,%%d0;0: bra.s 0b" : : "g"(cptr), "g"(plen) : "d0","cc");

        *ptr++ = 0x00;
        *ptr++ = 'L';
        *ptr++ = 'I';
        *ptr++ = 'B';
        *ptr++ = 0x00;
        *ptr++ = OTH_TAG;
    }

    SaveHeapUnLock();

    if (!(HeapRealloc(h,HeapSize(h)-120))) return MSG_ERR_TWINDEL; // Necessary ?

    if (twin.folder) {
        twin = SymFind(GenerateTIOSName(sbuffer,tmpname));
        if (!twin.folder) return MSG_ERR_TWINDEL;
        se = DerefSym(twin);
        if (!(se->flags.bits.twin)) return MSG_ERR_TWINDEL;
        se->flags.bits.hidden = 0;
        if (!SymDelTwin(se)) return MSG_ERR_TWINDEL;
    }

    if (EXCEPTION_TRIGGERED()) {
        plen = GET_EXCEPTION_CODE()+MSG_ERR_CRASHOFFSET;
        RESET_EXCEPTION();
        return plen;
    }
    else {
        return MSG_OKAY;
    }
}
#endif



/*===========================================================================*/
/* hides or unhides all files related to the explorer                        */
/*===========================================================================*/
void HideUnHideTICTFiles(short bhidden) {
    short          type;
    SYM_ENTRY*     tmp_entry;
    unsigned char* tmp_src;
    unsigned short tmp_len;

    tmp_entry = SymFindFirst(NULL,2);
    while (tmp_entry) {
        if (!tmp_entry->flags.bits.folder && tmp_entry->handle) {
            type = 0;
            if (!strncmp(tmp_entry->name,"tictex",8))        type = 1;
            else if (!strncmp(tmp_entry->name,"tictexpl",8)) type = 2;
            else {
                tmp_entry = SymFindNext();
                continue;
            }
            tmp_src = HeapDeref(tmp_entry->handle);
            tmp_len = *(unsigned short*)(tmp_src)+2;
            tmp_src += tmp_len-1;

            if ((type == 1 && *tmp_src == 0xF3)                   ||
                (type == 2 && !memcmp(tmp_src-4,"LIB\x00\xF8",5)))
            {
                tmp_entry->flags.bits.hidden = bhidden;
            }
        }
        tmp_entry = SymFindNext();
    }
}

/*===========================================================================*/
/* does cleanup of launcher if necessary (protected function)                */
/*===========================================================================*/
void HandleCleanup(short msg) {
    InstallCrashProtection(&temporary_protection);
    if (!ER_catch(temporary_errorframe)) {
        SaveHeapUnLock();
        HomeScreenRestore();
        if (stack_buffer) free(stack_buffer);
        stack_buffer = NULL;
        if (interface.lcdbuffer) free(interface.lcdbuffer);
        interface.lcdbuffer = NULL;
        SaveHeapUnLock();
        HideUnHideTICTFiles(0);
        top_estack = top_backup;
        FolderCur(GenerateTIOSName(initial_folder,tmpname),1);
        ER_success();
    }
    if (msg != MSG_OKAY) ST_helpMsg(GetMsg(msg));
    RemoveCrashProtection(&temporary_protection);
}


#define fatal(s) {ST_helpMsg(s); return;}


/*===========================================================================*/
/* MAIN ROUTINE: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    short          msg_number;
    char           temp_folder[30];
    SYM_ENTRY*     s_explorer = NULL;
    SYM_ENTRY*     tmp_entry;
    HANDLE         h_myself;
    HANDLE         tmp_h;
    unsigned char* tmp_src;
    unsigned short tmp_len;
    //unsigned char  firsttime;
    short          firsttime;

    unsigned long program_counter;

    //asm("0: bra.s 0b");
    if (TIOS_entries == 0x2AC) {
        //ST_helpMsg("AMS 1.00 is too old !");
        ((void(__ATTR_TIOS__ *)(const unsigned char *))((void *)(_rom_call_addr(E6))))("AMS 1.00 is too old !");
        return;
    }

    asm volatile("bsr 0f; 0:move.l (%%sp)+,%0":"=g"(program_counter));
    asm volatile("move.l %sp,stack_orig");

    // *Must* install this before calling any ROM_CALL using a F-Line instruction !
    InstallCrashProtection(&main_protection);

    SHOW_STACK(0);

    //-------------------------------------------------------------------------
    // initialize some variables ...
    //-------------------------------------------------------------------------
    memset((void*)&interface,0,sizeof(TICTEX_INTERFACE));
    SAVEHEAPLOCK_RESET();
    //lcdbuffer    = NULL;
    stack_buffer = NULL;
    top_backup   = top_estack;
    h_myself     = 0;
//    firsttime    = 0xFF;
    firsttime    = 1;
    FolderGetCur(initial_folder);

    //-------------------------------------------------------------------------
    // the correction value is evaluated by calling tststack() twice.
    // Once from the explorer and one time from the commandline
    //-------------------------------------------------------------------------
    stack_orig += STACK_CORRECT_VALUE;
    ((unsigned long)stack_orig) &= 0xfffffffe;

    //-------------------------------------------------------------------------
    // find myself and relocate myself if necessary (thanks Zeljko!)
    //-------------------------------------------------------------------------
    tmp_entry = SymFindFirst(NULL,2);
    while (tmp_entry) {
        tmp_h=tmp_entry->handle;
        if (tmp_h && HeapGetLock(tmp_h)) {
            tmp_src = HeapDeref(tmp_h);
            tmp_len = *(unsigned short*)tmp_src+2;
            if ((program_counter & 0x3FFFF) > (unsigned long)tmp_src &&
                (program_counter & 0x3FFFF) < (unsigned long)tmp_src+tmp_len-2)
            {
                h_myself  = tmp_h;
                break;
            }
        }
        tmp_entry = SymFindNext();
    }

    if (h_myself) {
// This test was incorrect in the first private beta of TICT-Explorer 1.40,
// leading to crashes on 92+/V200 without h220xTSR or HW2/3Patch...
        if (!(TI89 && ROM_base == (void *)0x800000)) {
            // Relocate self into the ghost space if not running on a 89T.
            if (program_counter<0x40000) {
                tmp_src = HeapDeref(h_myself);
                // I could use #define EXECUTE_IN_GHOST_SPACE, as reminded by
                // PpHd, but I'd need to call another TIGCCLIB function to get
                // h_myself, which would probably turn into bigger code.
                /*own_*/enter_ghost_space();
                EX_patch(tmp_src+0x40002,tmp_src + 0x40001 + *(unsigned short*)tmp_src);
            }
        }
    }
    else {
        ST_helpMsg(GetMsg(MSG_ERR_FINDMYSELF));
        RemoveCrashProtection(&main_protection); // Remove it now !
        return;
    }

    //-------------------------------------------------------------------------
    // Set my handle in the interface, so that tictexpl may use it to compute
    // the amount of memory the entire TICT-Explorer uses.
    //-------------------------------------------------------------------------
    interface.tictex_handle = h_myself;

    //-------------------------------------------------------------------------
    // allocate space which is used later to support programs with a full
    // 16kB stack (we need around 1500 bytes only)
    //-------------------------------------------------------------------------
    if (!(stack_buffer = (unsigned char*)malloc(STACK_MAX_SIZE))) {
        HandleCleanup(MSG_ERR_OUTOFMEM);
        RemoveCrashProtection(&main_protection); // Remove it now !
        return;
    }

    //InstallCrashProtection(&main_protection);
    if (!ER_catch(main_errorframe)) {

        //-------------------------------------------------------------------------
        // hide all files related to the explorer
        //-------------------------------------------------------------------------
        HideUnHideTICTFiles(1);

        //-------------------------------------------------------------------------
        // main loop
        //-------------------------------------------------------------------------
        do {
            interface.command = COMMAND_EXIT;

            //----------------------------------------------
            // find file tictexpl.LIB and start the explorer
            //----------------------------------------------
            explorer_folder[8] = 0;
            tmp_entry = SymFindFirst(NULL,2);
            while (tmp_entry) {
                if (tmp_entry->flags.bits.folder) {
                    strncpy(explorer_folder,tmp_entry->name,8);
                }
                else if (!strncmp(tmp_entry->name,"tictexpl",8)) {
                    if (!IsEntryInUse(tmp_entry)) {
                        tmp_src = HeapDeref(tmp_entry->handle);
                        tmp_len = *(unsigned short*)(tmp_src)+2;
                        tmp_src += tmp_len-1;

                        if (!memcmp(tmp_src-5,"\x00LIB\x00\xF8",6)) {
                            s_explorer = tmp_entry;
                            break;
                        }
                    }
                }
                tmp_entry = SymFindNext();
            }

            if (!s_explorer) {
                interface.message = MSG_ERR_EXPLLIBNOTFOUND;
                break;
            }

            //------------------------------------------------
            // allocate screen memory for the explorer and
            // start it (afterwards give the memory free again
            //------------------------------------------------
            if (!(interface.lcdbuffer = malloc(LCD_SIZE))) {
                 interface.message = MSG_ERR_OUTOFMEM;
                 break;
            }
            if (firsttime) {
                interface.command = COMMAND_FIRSTTIME;
                firsttime = 0;
            }
            else {
                interface.command = COMMAND_RUN;
            }

            interface.exemode = EXEMODE_NONE;
            msg_number = StartExecutable(s_explorer,1);

            free(interface.lcdbuffer);
            interface.lcdbuffer = NULL;

            if (msg_number) {
                interface.command = COMMAND_EXIT;
                interface.message = msg_number;
            }
            else if (interface.command != COMMAND_EXECUTE ||
                     interface.exemode == EXEMODE_BASIC)
            {
                break;
            }

            top_estack = top_backup;

            push_END_TAG();

            if (interface.exemode & EXEMODE_INTERPRET_FILE) {
                OSClearBreak();
                push_ANSI_string((char*)(interface.interpret_file));
            }
            else if (interface.exemode & EXEMODE_INTERPRET_FOLDERFILE) {
                sprintf(sbuffer,"%s\\%s",interface.interpret_folder,interface.interpret_file);
                OSClearBreak();
                push_ANSI_string(sbuffer);
            }

            sprintf(sbuffer,"%s\\%s",interface.foldername,interface.filename);
            tmp_entry = GetSymByName(sbuffer);

            if (!tmp_entry || !tmp_entry->handle) {
                interface.message = MSG_ERR_FILENOTFOUND;
            }
            else if (IsEntryInUse(tmp_entry)) {
                interface.message = MSG_ERR_FILEINUSE;
            }
            else {
                FolderGetCur(temp_folder);
                FolderCur(GenerateTIOSName((char*)interface.foldername,tmpname),1);
                if (interface.exemode & EXEMODE_PPG) {
                    interface.message = StartPPG(tmp_entry->handle);
                }
                else {
                    interface.message = StartExecutable(tmp_entry,0);
                }
                // restore old folder again
                FolderCur(GenerateTIOSName(temp_folder,tmpname),1);
            }

            //---------------------------------------------------------------------
            // if a program leaves some keypresses in the queue -> remove them and
            // restore the graphics port as well
            //---------------------------------------------------------------------
            PortRestore();
            GKeyFlush();

            top_estack = top_backup;
        }
        while (interface.command != COMMAND_EXIT);
        ER_success();
    }
    //RemoveCrashProtection(&main_protection);

    //---------------------------------------------------------------------
    // is something triggered by myself ??? (should never happen)
    //---------------------------------------------------------------------
    if (EXCEPTION_TRIGGERED()) {
        HandleCleanup(GET_EXCEPTION_CODE()+MSG_ERR_CRASHOFFSET);
        RemoveCrashProtection(&main_protection); // Remove it now !
        return;
    }

    HandleCleanup(MSG_OKAY);
    //------------------------------------------------------------------------
    // paste BASIC program string if wanted
    //------------------------------------------------------------------------
    if (interface.exemode == EXEMODE_BASIC) {
        EVENT ev;
        HANDLE h;
        sprintf(sbuffer,"%s\\%s()",interface.foldername,interface.filename);
        CB_replaceTEXT(sbuffer, strlen(sbuffer), 0);
        ev.Type = CM_CLR;
        EV_sendEvent(h = EV_getAppID("TIHOME"),&ev);
        ev.Type = CM_MENU_PASTE;
        EV_sendEvent(h/*EV_getAppID("TIHOME")*/,&ev);
    }

    // Err, why is this line duplicated ??
    if (interface.message) ST_helpMsg(GetMsg(interface.message));
    if (interface.message) ST_helpMsg(GetMsg(interface.message));

    OSClearBreak();        // clear the break key for sure ...
    GKeyFlush();
//    OSqclear(kbd_queue()); // clear keyboard queue for sure ..
    RemoveCrashProtection(&main_protection); // Remove it now !
}


//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: tictex.c,v $
// Revision 1.17  2002/10/29 09:07:25  tnussb
// workaround for the TWIN symbol deletion bug of the AMS: sets now the
// inuse-bit (hidden-bit) if we generate a TWIN symbol for executing an
// archived program. Thanx for Samuel Stearley who reported this AMS bug
// first.
//
// Revision 1.16  2002/09/10 11:28:42  tnussb
// changes up to v1.30 Beta 4 / examine history.txt for details
//
// Revision 1.15  2002/03/15 15:11:25  tnussb
// (1) necessary modifications for V200 (EX_patch fix etc.)
// (2) OSClearBreak() and OSqclear() added to end of program
//
// Revision 1.14  2002/02/21 09:31:23  tnussb
// VERSION_TI89 and VERSION_TI92P replaced by new standard USE_TI89 and USE_TI92P
//
// Revision 1.13  2002/02/07 18:01:18  tnussb
// generic commit
//
// Revision 1.12  2001/02/04 13:11:28  Thomas Nussbaumer
// changes up to version 1.00 RC2 (see history.txt)
//
// Revision 1.11  2001/01/26 21:04:44  Thomas Nussbaumer
// changes for version 0.80 [see history.txt]
//
// Revision 1.10  2001/01/21 17:05:09  Thomas Nussbaumer
// changes related to v0.65
//
// Revision 1.9  2001/01/21 00:20:16  Thomas Nussbaumer
// additionally changes related to version 0.60
//
// Revision 1.8  2001/01/20 21:36:30  Thomas Nussbaumer
// changes due to version 0.60
//
// Revision 1.7  2001/01/17 23:39:22  Thomas Nussbaumer
// (1) moved complete source to main.c
// (2) implemented here a custom startup utility
//
//
